/* standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <strings.h>
#include <time.h>

/* library headers */
#include "cgi-util.h"

/* program headers */
#include "flexi-cgi.h"
#include "actions.h"
#include "tree.h"

static action_t actions[]=
{
    {acc_maint,"update account"},
    {list_weeks,"list"},
    {list_leave,"leave"},
    {list_summaries,"summary"},
    {list_general_summary,"general"},
    {list_4week_summary,"four week"},
    {list_full_summary,"full"},
    {NULL,NULL}
};


int action(fx_week * thisweek,const char * docroot,const char * user,const char * password,const char * action)
{
    int loop;
    int match;

    /* search command list for match */
    for(loop=0;actions[loop].proc != NULL;loop++)
    {
	match=check_match(action,actions[loop].actn);

	switch(match)
	{
	case 0: /* good match */
	    return (actions[loop].proc)(thisweek,docroot,user,password);
	}
    }

    /* no matches in list so default action */
    return week_input(thisweek,docroot,user,password);

}


int check_match(const char * string,char * cmnd)
{
    int str_len;
    int cmnd_len;

    if(string==NULL)
	return 1;

    str_len=strlen(string);
    cmnd_len=strlen(cmnd);

    /* check for exact match on non arument command */
    if((str_len == cmnd_len) &&
       (strncasecmp(string,cmnd,cmnd_len)==0))
	   return 0; /* good match */

    return 1; /* no match */
}



/*
 * support funcs for actions
 */


int do_table_cell(const time_t * toshow,const char * dayname,const char * cell_name)
{
    struct tm *tm_time;
    tm_time=gmtime(toshow);

    printf("<td><input type=\"text\" name=\"%s_h_%s\" value=\"%.2d\" maxlength=2 size=3>",dayname,cell_name, tm_time->tm_hour);

    printf(":<input type=\"text\" name=\"%s_m_%s\" value=\"%.2d\" maxlength=2 size=3>\n",dayname,cell_name,tm_time->tm_min);
    return 0;
}

int do_table_cell_total(const time_t toshow)
{
    int hours,mins;
    hours=(toshow)/3600;
    mins=((toshow)-(hours*3600))/60;
    if((hours<0) || (mins<0))
    {
	/* deal with negative time printing nicely */
	hours=0-hours;
	mins=0-mins;
    printf("<td><h4>-%.2d:%.2d</h4>", hours,mins);
    }
    else
    printf("<td><h4>%.2d:%.2d</h4>", hours,mins);
    return 0;
}

int do_row_title(const fx_day *day)
{
    char * text;

    switch(day->daytype)
    {
    case DAYTYPE_BANKHOL:
	text=" is a Bank Holiday!";
	break;

    case DAYTYPE_HALF_HOLIDAY:
	text=" taken as half holiday";
	break;

    case DAYTYPE_HALF_FLEXIDAY:
	text=" taken as a half flexi day";
	break;

    case DAYTYPE_HALF_SICK:
	text=" was a half sick day";
	break;

    case DAYTYPE_HOLIDAY:
	text=" taken as holiday";
	break;

    case DAYTYPE_FLEXIDAY:
	text=" taken as a flexi day";
	break;

    case DAYTYPE_SICK:
	text=" was a sick day";
	break;

    case DAYTYPE_CONFERENCE:
	text=" was a conference day";
	break;

    default:
	text="";
	break;

    }

    printf("<th>%s%s</th>",day->dayname, text);
    return 0;
}

int do_row_entries(const fx_day *day)
{
    switch(day->daytype)
    {
    case DAYTYPE_BANKHOL:
	do_table_cell(&day->inout[0],day->dayname,"inout_0");
	do_table_cell(&day->inout[1],day->dayname,"inout_1");
	do_table_cell(&day->inout[2],day->dayname,"inout_2");
	do_table_cell(&day->inout[3],day->dayname,"inout_3");
	do_table_cell_total(day->day_total);
	break;

    case DAYTYPE_HALF_HOLIDAY:
    case DAYTYPE_HALF_FLEXIDAY:
    case DAYTYPE_HALF_SICK:
	do_table_cell(&day->inout[0],day->dayname,"inout_0");
	do_table_cell(&day->inout[1],day->dayname,"inout_1");
	do_table_cell_total(day->day_total);
	break;

    case DAYTYPE_HOLIDAY:
    case DAYTYPE_FLEXIDAY:
    case DAYTYPE_SICK:
    case DAYTYPE_CONFERENCE:
	break;

    default:
	do_table_cell(&day->inout[0],day->dayname,"inout_0");
	do_table_cell(&day->inout[1],day->dayname,"inout_1");
	do_table_cell(&day->inout[2],day->dayname,"inout_2");
	do_table_cell(&day->inout[3],day->dayname,"inout_3");
	do_table_cell_total(day->day_total);
    }

    return 0;
}

int do_table_day(const fx_day *day)
{
    printf("<tr>");
    do_row_title(day);
    do_row_entries(day);
    printf("</tr>\n");
    return 0;
}


/*
 * action functions
 */

int week_input(fx_week * thisweek,const char * docroot,const char * user,const char * password)
{
    int loop;

    printf("<head><title>Flexi Time Sheet For %s</title></head><body>",user);

    printf("<h2>Flexi Time Sheet For %s</h2>",user);

    printf("<h3>Today is %s in week number %d year %d</h3>",thisweek->days[thisweek->dayno].dayname,thisweek->weekno,thisweek->yearno);

    printf("<form action=\"/cgi-bin/flexi-cgi\" method=\"post\">\n");

    printf("<input type=\"hidden\" name=\"user\" value=\"%s\">\n",user);

    if(password!=NULL)
	printf("<input type=\"hidden\" name=\"password\" value=\"%s\">\n",password);
    printf("<input type=\"hidden\" name=\"year\" value=\"%d\">\n",thisweek->yearno);
    printf("<input type=\"hidden\" name=\"week\" value=\"%d\">\n",thisweek->weekno);

    printf("<table cellspacing=6>");
    printf("<tr><th><th colspan=5>Times</tr>\n");
    printf("<tr><th><th>In<th>Out<th>In<th>Out<th>Total</tr>\n");

    for (loop=0;loop<7;loop++)
    {
	do_table_day(&thisweek->days[loop]);
    }

    printf("<tr><td><td><td><td><td><h4>Hours Required</h4></td>");
    do_table_cell_total(thisweek->week_length);
    printf("</tr><tr><td><td><td><td><td><h4>Total Worked</h4></td>");
    do_table_cell_total(thisweek->week_total);
    printf("</tr><tr><td><td><td><td><td><h4>Total Leave</h4></td>");
    do_table_cell_total(thisweek->week_leave);
    printf("</tr><tr><td><td><td><td><td><h4>Total Excess</h4></td>");
    do_table_cell_total((thisweek->week_total + thisweek->week_leave) - thisweek->week_length);
    printf("</tr>");

    printf("</table>");
    printf("<hr><input type=\"submit\" name=\"action\" value=\"Update\">\n");
    printf("<input type=\"submit\" name=\"action\" value=\"List\">\n");
    printf("<input type=\"submit\" name=\"action\" value=\"Leave\">\n");
    printf("<input type=\"submit\" name=\"action\" value=\"Summary\">\n");
    printf("<input type=\"submit\" name=\"action\" value=\"Update Account\">\n");
    printf("</form></body>\n");
    return 0;
}

int list_leave(fx_week * listweek,const char * docroot,const char * user,const char * password)
{
    fx_day * day;
    int loop;
    int typeloop;

    printf("<head><title>Leave Sheet For %s</title></head><body>",user);

    printf("<h2>Leave Sheet For %s</h2>",user);

    printf("<h3>Today is %s in week number %d</h3>",listweek->days[listweek->dayno].dayname,listweek->weekno);

    printf("<form action=\"/cgi-bin/flexi-cgi\" method=\"post\">\n");

    printf("<input type=hidden name=\"user\" value=\"%s\">\n",user);

    if(password!=NULL)
	printf("<input type=hidden name=\"password\" value=\"%s\">\n",password);
    printf("<input type=hidden name=\"year\" value=\"%d\">\n",listweek->yearno);
    printf("<input type=hidden name=\"week\" value=\"%d\">\n",listweek->weekno);

    printf("<table cellspacing=6>\n");
    printf("<tr><th><th>Normal<th>Sick<th>Holiday<th>Conference<th>Flexi Day<th>Overtime<th>Half Holiday<th>Half Flexi<th>Half Sick</tr>\n");

    for (loop=0;loop<7;loop++)
    {
	day = &listweek->days[loop];
	printf("<tr><td>%s",day->dayname);

	for(typeloop=DAYTYPE_GLOBALEND;typeloop<=DAYTYPE_USEREND;typeloop++)
	{
	    printf("<td><input type=radio");
	    if (typeloop>DAYTYPE_GLOBALEND)
	    {
		if(day->daytype==typeloop)
		    printf(" checked");
	    }
	    else
	    {
		if(day->daytype<=DAYTYPE_GLOBALEND)
		    printf(" checked");
	    }
		printf(" name=\"%s_daytype\" value=\"%.3d\">",
		       day->dayname,
		       typeloop );


	}
	printf("</tr>\n");

    }
    printf("</table>");

    printf("<hr><input type=submit name=\"action\" value=\"Update\">\n");
    printf("</form>\n");
    return 0;
}

int acc_maint(fx_week * thisweek,const char * docroot,const char * user,const char * password)
{
    int pwdupd;

    pwdupd=flexi_update_passwd(docroot,user);

    printf("<h2>Account Maintinance</h2>");

    switch(pwdupd)
    {
    case 1:
	printf("<h1>Sucessfuly updated Password</h1>");
	break;
    case 2:
	printf("<h1>Old Password Did Not Match</h1>");
	break;
    case 3:
	printf("<h1>New Password Not Valid</h1>");
	break;
    case 4:
	printf("<h1>New Passwords Did Not Match</h1>");
	break;
    case 5:
	printf("<h1>System Error Setting Password</h1>");
	break;
    }

    printf("<form action=\"/cgi-bin/flexi-cgi\" method=\"post\">");

    if(pwdupd==1)
	printf("<input type=\"hidden\" name=\"password\" value=\"%s\">\n",cgi_getentrystr("newpassword1"));
    else if(password!=NULL)
	printf("<input type=\"hidden\" name=\"password\" value=\"%s\">\n",password);

    printf("<input type=\"hidden\" name=\"user\" value=\"%s\">\n",user);

    printf("<input type=\"hidden\" name=\"year\" value=\"%d\">\n",thisweek->yearno);
    printf("<input type=hidden name=\"week\" value=\"%d\">\n",thisweek->weekno);

    printf("<table>\n");

    printf("<tr><th colspan=2>Password</th></tr>");

    printf("<tr><td>Old Password</td><td><input type=\"password\" name=\"oldpassword\" value=\"\"></tr>\n");
    printf("<tr><td>New Password</td><td><input type=\"password\" name=\"newpassword1\" value=\"\"></tr>\n");
    printf("<tr><td>Repeat New Password</td><td><input type=\"password\" name=\"newpassword2\" value=\"\"></tr>\n");
    printf("</table>\n");


    printf("<hr><input type=submit name=\"action\" value=\"Update Account\">\n");
    printf("<hr><input type=submit name=\"action\" value=\"Done\">\n");
    printf("</form>\n");
    return 0;

}


int walk_tree_list(const char * user,
		   const char * password,
		   int selweekno,
		   int *yearno,
		   struct treenode* node)
{

    if (node->left != NULL)
	walk_tree_list(user,password,selweekno,yearno,node->left);

    if(*yearno != node->yearno)
    {
	if(*yearno!=0)
	{
	    printf("</select>");
	    printf("<input type=\"submit\" name=\"action\" value=\"Select\"><hr>\n");
	    printf("</form>");
	}

	*yearno=node->yearno;

	printf("<form action=\"/cgi-bin/flexi-cgi\" method=\"post\">");

	printf("<input type=\"hidden\" name=\"user\" value=\"%s\">\n",user);

	printf("<input type=\"hidden\" name=\"year\" value=\"%d\">\n",*yearno);

	if(password!=NULL)
	    printf("<input type=\"hidden\" name=\"password\" value=\"%s\">\n",
		   password);

	printf("<h2>Year %04d</h2>\n",*yearno);

	printf("<select name=\"week\">");

    }

    if(node->weekno == selweekno )
	    printf("<option selected>%d",node->weekno);
	else
	    printf("<option>%d",node->weekno);


    if (node->right != NULL)
	walk_tree_list(user,password,selweekno,yearno,node->right);

    return 0;
}

/* list previous weeks to select from */
int list_weeks(fx_week * thisweek,
	       const char * docroot,
	       const char * user,
	       const char * password)
{
    char * user_dir_name;
    struct treenode * new_treeroot;

    int yearno=0;

    user_dir_name=malloc(strlen(docroot)+strlen(FLEXI_CGI_ROOT)+strlen(user)+3);

    strcpy(user_dir_name,docroot);
    strcat(user_dir_name,"/"FLEXI_CGI_ROOT"/");
    strcat(user_dir_name,user);

    new_treeroot = make_tree_list(user_dir_name);

    free(user_dir_name);

    if(new_treeroot == NULL)
    {
	send_error("Unable to access users sheet directory");
	return -1;
    }

    printf("<h1>Available Weeks by year</h1>");

    walk_tree_list(user,password,thisweek->weekno,&yearno,new_treeroot);

    printf("</select>");
    printf("<input type=\"submit\" name=\"action\" value=\"Select\">\n");
    printf("<hr><input type=submit name=\"action\" value=\"Done\">\n");
    printf("</form>");

    free_tree_list(new_treeroot);

    return 0;
}

int list_summaries(fx_week * listweek,const char * docroot,const char * user,const char * password)
{
    printf("<h2>Available Summaries</h2>");
    printf("<form action=\"/cgi-bin/flexi-cgi\" method=\"post\">\n");

    printf("<input type=hidden name=\"user\" value=\"%s\">\n",user);

    if(password!=NULL)
	printf("<input type=hidden name=\"password\" value=\"%s\">\n",password);
    printf("<input type=hidden name=\"year\" value=\"%d\">\n",listweek->yearno);
    printf("<input type=hidden name=\"week\" value=\"%d\">\n",listweek->weekno);

    printf("<p><input type=submit name=\"action\" value=\"General\">\n");
    printf("<p><input type=submit name=\"action\" value=\"Four Week\">\n");
    printf("<p><input type=submit name=\"action\" value=\"Full\">\n");
    printf("<hr><input type=submit name=\"action\" value=\"Done\">\n");
    printf("</form>\n");
    return 0;

}

int list_general_summary(fx_week * listweek,const char * docroot,const char * user,const char * password)
{
    fx_week * sumweek;
    int year=0;
    int start_week=0;
    int last_week=0;
    int four_week=0;
    int loop;
    int num_weeks=0;

    long total_worked=0;
    long total_required=0;
    long total_leave=0;
    long total_flexi=0;

    long total_4_worked=0;
    long total_4_required=0;
    long total_4_leave=0;
    long total_4_flexi=0;

    year=get_year_number();
    last_week=get_week_number();
    four_week=last_week-4;

    printf("<head><title>General Summary Report For %s</title></head><body>",user);
    printf("<h2>General Summary Report For %s</h2>",user);
    printf("<h1>Report from week %d to week %d in %04d</h1>\n",start_week,last_week,year);


    /* doesnt include this week */
    for (loop=start_week;loop<last_week;loop++)
    {
	sumweek=load_week(docroot,user,loop,year);

	if (sumweek!=NULL)
	{
	    calc_worked_hours(sumweek);
	    num_weeks++;
	    total_worked+=sumweek->week_total;
	    total_required+=sumweek->week_length;
	    total_leave+=sumweek->week_leave;
	    total_flexi+=sumweek->week_flexi;

	    if (loop>=four_week)
	    {
		total_4_worked+=sumweek->week_total;
		total_4_required+=sumweek->week_length;
		total_4_leave+=sumweek->week_leave;
		total_4_flexi+=sumweek->week_flexi;
	    }

	    free(sumweek);
	}
    }


    printf("<table border=""1"">");

    printf("<tr><th>General Statistics</th></tr>");

    printf("<tr><td>Complete records exist for a total of %d weeks</td></tr>",num_weeks);
    printf("<tr><td>Totals for weeks %d to %d</td></tr>",(last_week-num_weeks+1),last_week);

    printf("<tr><td>Required hours</td>");
    do_table_cell_total(total_required);
    printf("</tr>\n");

    printf("<tr><td>Worked hours</td>");
    do_table_cell_total(total_worked);
    printf("</tr>\n");

    printf("<tr><td>Leave hours</td>");
    do_table_cell_total(total_leave);
    printf("</tr>\n");

    printf("</tr><tr><td>Flexi Converted</td>");
    do_table_cell_total(-total_flexi);
    printf("</tr>\n");

    printf("<tr><td>Excess hours</td>");
    do_table_cell_total((total_worked + total_leave) - (total_required + total_flexi));
    printf("</tr>\n");

    printf("<tr><th>Totals for last Four weeks</th></tr>");

    printf("<tr><td>Required hours");
    do_table_cell_total(total_4_required);

    printf("</tr><tr><td>Worked hours");
    do_table_cell_total(total_4_worked);

    printf("</tr><tr><td>Leave hours</td>");
    do_table_cell_total(total_4_leave);
    printf("</tr>\n");

    printf("</tr><tr><td>Flexi Converted</td>");
    do_table_cell_total(-total_4_flexi);
    printf("</tr>\n");

    printf("<tr><td>Excess");
    do_table_cell_total((total_4_worked + total_4_leave)-(total_4_required + total_4_flexi));

    printf("</tr></table><hr>");


    printf("<form action=\"/cgi-bin/flexi-cgi\" method=\"post\">\n");

    printf("<input type=hidden name=\"user\" value=\"%s\">\n",user);

    if(password!=NULL)
	printf("<input type=hidden name=\"password\" value=\"%s\">\n",password);
    printf("<input type=hidden name=\"year\" value=\"%d\">\n",listweek->yearno);
    printf("<input type=hidden name=\"week\" value=\"%d\">\n",listweek->weekno);

    printf("<input type=submit name=\"action\" value=\"Summary\">\n");
    printf("<input type=submit name=\"action\" value=\"Done\">\n");
    printf("</form>\n");

    return 0;
}

int list_4week_summary(fx_week * listweek,const char * docroot,const char * user,const char * password)
{
    fx_week * sumweek;
    int start_week;
    int last_week=0;
    int dayloop;
    int loop;

    int year;

    long total_worked=0;
    long total_required=0;
    long total_leave=0;
    long total_flexi=0;

    year=get_year_number();
    last_week=get_week_number();
    start_week=last_week-4;

    printf("<head><title>Four Week Report For %s</title></head><body>",user);
    printf("<h2>Four Week Report For %s</h2>\n",user);
    printf("<h1>Report from week %d to week %d in %04d</h1>\n",start_week,last_week,year);

    for (loop=start_week;loop<last_week;loop++)
    {

	sumweek=load_week(docroot,user,loop,year);

	if (sumweek!=NULL)
	{
	    printf("<table border=""1"">");
	    printf("<tr><th colspan=3>Week Number %d</th></tr>",loop);
	    printf("<tr><th>Day</th><th>Worked</th><th>Excess</th></tr>");

	    calc_worked_hours(sumweek);
	    total_worked+=sumweek->week_total;
	    total_required+=sumweek->week_length;
	    total_leave+=sumweek->week_leave;
	    total_flexi+=sumweek->week_flexi;

	    for (dayloop=0;dayloop<7;dayloop++)
	    {

		/* skip empty days */
		if(sumweek->days[dayloop].day_total==0)
		    continue;

		printf("<tr>");
		do_row_title(&sumweek->days[dayloop]);
		do_table_cell_total(sumweek->days[dayloop].day_total);


		do_table_cell_total(sumweek->days[dayloop].day_total - sumweek->work_day_length);
		printf("</tr>\n");
	    }

	    printf("<tr><th colspan=2>Total Required</th>");
	    do_table_cell_total(sumweek->week_length);
	    printf("</tr>\n");

	    printf("<tr><th colspan=2>Total Worked</th>");
	    do_table_cell_total(sumweek->week_total);
	    printf("</tr>\n");

	    printf("<tr><th colspan=2>Total Leave</th>");
	    do_table_cell_total(sumweek->week_leave);
	    printf("</tr>\n");

	    printf("<tr><th colspan=2>Flexi Converted</th>");
	    do_table_cell_total(-sumweek->week_flexi);
	    printf("</tr>\n");

	    printf("<tr><th colspan=2>Total Excess</th>");
	    do_table_cell_total((sumweek->week_total + sumweek->week_leave) - sumweek->week_length + sumweek->week_flexi );
	    printf("</tr>\n");

	    free(sumweek);

	    printf("</table>\n");

	}
    }

    printf("<h2>Totals For Last 4 weeks</h2>\n");

    printf("<table><tr><td>Total Required</td>");
    do_table_cell_total(total_required);
    printf("</tr><tr><td>Total Worked</td>");
    do_table_cell_total(total_worked);
    printf("</tr><tr><td>Total Leave</td>");
    do_table_cell_total(total_leave);
    printf("</tr><tr><td>Total Flexi Converted</td>");
    do_table_cell_total(-total_flexi);
    printf("</tr><tr><td>Total Excess</td>");
    do_table_cell_total((total_worked + total_leave)-(total_required + total_flexi));
    printf("</tr></table>");

    printf("<form action=\"/cgi-bin/flexi-cgi\" method=\"post\">\n");

    printf("<input type=hidden name=\"user\" value=\"%s\">\n",user);

    if(password!=NULL)
	printf("<input type=hidden name=\"password\" value=\"%s\">\n",password);
    printf("<input type=hidden name=\"year\" value=\"%d\">\n",listweek->yearno);
    printf("<input type=hidden name=\"week\" value=\"%d\">\n",listweek->weekno);
    printf("<hr><input type=\"submit\" name=\"action\" value=\"Summary\">\n");
    printf("<input type=submit name=\"action\" value=\"Done\">\n");
    printf("</form>\n");

    return 0;

}

int list_full_summary(fx_week * listweek,const char * docroot,const char * user,const char * password)
{
    fx_week * sumweek;
    int start_week;
    int last_week=0;
    int loop;
    int year;

    year=get_year_number();
    last_week=get_week_number();
    start_week=0;

    printf("<h2>Full Report</h2>\n");
    printf("<h1>Report from week %d to week %d in %04d</h1>\n",start_week,last_week,year);

    printf("<table border=""1""><tr><th>Week Number</th><th>Required</th><th>Worked</th><th>Leave</th><th>Excess</th></tr>");
    for (loop=start_week;loop<last_week;loop++)
    {
	sumweek=load_week(docroot,user,loop,year);

	if (sumweek!=NULL)
	{
	    calc_worked_hours(sumweek);
	    printf("<tr><td>%d</td>",loop);
	    do_table_cell_total(sumweek->week_length);
	    do_table_cell_total(sumweek->week_total);
	    do_table_cell_total(sumweek->week_leave);
	    do_table_cell_total((sumweek->week_total + sumweek->week_leave) - sumweek->week_length );
	    printf("</tr>\n");
	    free(sumweek);
	}
    }

    printf("</table>");

    printf("<form action=\"/cgi-bin/flexi-cgi\" method=\"post\">\n");

    printf("<input type=hidden name=\"user\" value=\"%s\">\n",user);

    if(password!=NULL)
	printf("<input type=hidden name=\"password\" value=\"%s\">\n",password);
    printf("<input type=hidden name=\"year\" value=\"%d\">\n",listweek->yearno);
    printf("<input type=hidden name=\"week\" value=\"%d\">\n",listweek->weekno);
    printf("<hr><input type=submit name=\"action\" value=\"Summary\">\n");
    printf("<input type=submit name=\"action\" value=\"Done\">\n");
    printf("</form>\n");

    return 0;

}
