/*
 * flexi-cgi is a simple flexitime cgi recorder
 * Copyright (C) 2000 V.R.Sanders
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * To contact V Sanders email
 * vince@kyllikki.org
 *
 * The main web site for this program is
 * http://www.kyllikki.org/software
 *
 * flexi-cgi uses cgi-util LGPL library by Bill Kendrick (see cgi-util directory)
 */

/* standard headers */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <strings.h>
#include <time.h>

/* library headers */
#include "cgi-util.h"

/* program headers */
#include "flexi-cgi.h"
#include "actions.h"

int send_error(const char * err_txt)
{
    printf("<head>\n<title>Error :%s</title>\n</head><body>\n<h1>Error :%s</h1>\n</body>\n",err_txt,err_txt);
    return 0;
}

int flexi_user_list(const char * docroot)
{
    char * user_dir_name;
    DIR * user_dir;
    struct dirent * user_dir_entry;
    int first=1;

    user_dir_name=malloc(strlen(docroot)+strlen(FLEXI_CGI_ROOT)+2);

    strcpy(user_dir_name,docroot);
    strcat(user_dir_name,"/"FLEXI_CGI_ROOT);

    user_dir=opendir(user_dir_name);

    if (user_dir==NULL)
    {
	send_error("Unable to access flexi sheets directory");
	return -1;
    }

    printf("<h2>Select User</h2>");
    printf("<form action=\"/cgi-bin/flexi-cgi\" method=\"post\">");

    printf("<select name=\"user\">");

    while((user_dir_entry=readdir(user_dir)))
    {
	/* skip hidden files */
	if (user_dir_entry->d_name[0]=='.')
	    continue;
	if(first)
	    printf("<option selected>%s",
		   user_dir_entry->d_name);
	else
	    printf("<option>%s",
		   user_dir_entry->d_name);
    }

    printf("</select>");
    printf("<hr><input type=\"submit\" name=\"action\" value=\"Select\">\n");
    printf("</form>");

    closedir(user_dir);
    free(user_dir_name);
    return 0;
}

int flexi_set_passwd(const char * docroot,const char * user,const char * password)
{
    char * paswd_file_name;
    FILE * paswd_file;

    if (user==NULL)
	return 0;

    paswd_file_name=malloc(strlen(docroot)+strlen(FLEXI_CGI_ROOT)+strlen(user)+20);

    strcpy(paswd_file_name,docroot);
    strcat(paswd_file_name,"/"FLEXI_CGI_ROOT"/");
    strcat(paswd_file_name,user);
    strcat(paswd_file_name,"/.paswd");

    paswd_file=fopen(paswd_file_name,"w");

    if (paswd_file==NULL)
	return 1;

    fwrite( password, strlen(password), 1, paswd_file);

    fclose(paswd_file);

    free(paswd_file_name);

    return 0;
}


int flexi_check_passwd(const char * docroot,const char * user,const char * password)
{
    char * paswd_file_name;
    FILE * paswd_file;
    char * paswd_read;
    int result=0;
    int loop=0;

    if (user==NULL)
	return 0;

    paswd_file_name=malloc(strlen(docroot)+strlen(FLEXI_CGI_ROOT)+strlen(user)+20);

    strcpy(paswd_file_name,docroot);
    strcat(paswd_file_name,"/"FLEXI_CGI_ROOT"/");
    strcat(paswd_file_name,user);
    strcat(paswd_file_name,"/.paswd");

    paswd_file=fopen(paswd_file_name,"r");

    /* no file? no passwd allow acess - dodgy */
    if (paswd_file==NULL)
	return 1;

    if (password!=NULL)
    {

	paswd_read=malloc(1024);
	paswd_read[fread(paswd_read,1,1024,paswd_file)]=0;

	while(paswd_read[loop])
	{
	    if(paswd_read[loop]<32)
		paswd_read[loop]=0;
	    loop++;
	}


	if(strcmp(paswd_read,password)==0)
	    result=1;

	free(paswd_read);
    }

    fclose(paswd_file);

    free(paswd_file_name);


    return result;
}

int flexi_update_passwd(const char * docroot,const char * user)
{
    const char * oldpass;
    const char * newpass1;
    const char * newpass2;

    /* bail if theres no user */
    if (user==NULL)
	return 0;

    oldpass = cgi_getentrystr("oldpassword");
    if(oldpass==NULL)
	return 0; /* unable to get old password */

    if(flexi_check_passwd(docroot,user,oldpass)==0)
	return 2;

    newpass1 = cgi_getentrystr("newpassword1");
    if(newpass1==NULL)
	return 3; /* bad pw change */

    newpass2 = cgi_getentrystr("newpassword2");
    if(newpass2==NULL)
	return 3; /* bad pw change */

    if(strcmp(newpass1,newpass2)!=0)
	return 4; /* new didnt match */

    if(flexi_set_passwd(docroot,user,newpass1))
	return 5;

    return 1;

}


int flexi_prompt_passwd(const char * docroot,const char * user)
{
    if (user==NULL)
    {
	/* no user specified so list available ones */
     	flexi_user_list(docroot);
	return 0;
    }

    printf("<h2>Password Required</h2>\n");
    printf("<form action=\"/cgi-bin/flexi-cgi\" method=\"post\">\n");

    printf("<input type=\"hidden\" name=\"user\" value=\"%s\">\n",user);

    printf("Password required for user %s",user);

    printf("<p>Password <input type=\"password\" name=\"password\">\n");

    printf("<hr><input type=\"submit\" name=\"action\" value=\"Select\">\n");
    printf("</form>\n");

    return 0;
}

int get_week_number()
{
    struct tm* tm_timenow;
    char week_num[128];
    time_t timenow;

    timenow=time(NULL);
    tm_timenow=gmtime(&timenow);
    strftime(week_num,128,"%W",tm_timenow);
    return(atoi(week_num));

}

int get_year_number()
{
    struct tm* tm_timenow;
    char year_num[128];
    time_t timenow;

    timenow=time(NULL);
    tm_timenow=gmtime(&timenow);
    strftime(year_num,128,"%Y",tm_timenow);
    return(atoi(year_num));
}

/* find the weekday number 0 based from monday */
int get_week_day_number()
{
    struct tm* tm_timenow;
    time_t timenow;
    int day;

    timenow=time(NULL);
    tm_timenow=gmtime(&timenow);
    day=tm_timenow->tm_wday;
    day--;
    if(day==-1)
	day=6;
    return(day);
}


FILE * open_sheet(const char * docroot,const char * user,int year_num,int week_num,char* mode)
{
    FILE * sheet_file;
    char * sheet_file_name;

    if (user==NULL)
	return NULL;

    sheet_file_name=malloc(strlen(docroot)+strlen(FLEXI_CGI_ROOT)+strlen(user)+24);

    sprintf(sheet_file_name,"%s/%s/%s/%04d-week-%d",docroot,FLEXI_CGI_ROOT,user,year_num,week_num);

    sheet_file=fopen(sheet_file_name,mode);

    if (sheet_file==NULL)
    {
	if(mode[0]=='r')
	{
	    /* backward compat - if cant open file with year try opening
	       without year reference - safe coz we will never save as this
	       filename - may just load last years data into this year */
	    sprintf(sheet_file_name,"%s/%s/%s/week-%d",docroot,FLEXI_CGI_ROOT,user,week_num);

	    sheet_file=fopen(sheet_file_name,mode);

	    if (sheet_file==NULL)
	    {
		free(sheet_file_name);
		return NULL;
	    }
	}
	else
	{
	    free(sheet_file_name);
	    return NULL;
	}

    }

    free(sheet_file_name);

    return sheet_file;
}

int calc_normal_hours(fx_week * calcweek)
{
    int loop;
    int total=0;

    for(loop=0;loop<7;loop++)
    {
	if((calcweek->days[loop].daytype!=DAYTYPE_WEEKEND) &&
	   (calcweek->days[loop].daytype!=DAYTYPE_BANKHOL))
	    total+=calcweek->work_day_length;
    }
    return total;
}



int read_week_control(const char * docroot,fx_week * calcweek)
{
    FILE * ctrl_file;

    ctrl_file=open_sheet(docroot,".control",calcweek->yearno,calcweek->weekno,"r");

    if (ctrl_file==NULL)
    {
	/* use defaults if cannot open control file */
	calcweek->work_day_length=27000; /* 7.5 hours */
	calcweek->days[5].daytype_default=DAYTYPE_WEEKEND;
	calcweek->days[6].daytype_default=DAYTYPE_WEEKEND;
    }
    else
    {
	fscanf(ctrl_file,
	       "%ld %d %d %d %d %d %d %d",
	       &calcweek->work_day_length,
	       &calcweek->days[0].daytype_default,
	       &calcweek->days[1].daytype_default,
	       &calcweek->days[2].daytype_default,
	       &calcweek->days[3].daytype_default,
	       &calcweek->days[4].daytype_default,
	       &calcweek->days[5].daytype_default,
	       &calcweek->days[6].daytype_default);

    fclose(ctrl_file);

    }

    return 0;
}

fx_week* initweek(const char * docroot,int weeknumber,int yearnumber)
{
    fx_week * toset;
    int loop;

    toset=malloc(sizeof(fx_week));
    if (toset==NULL)
	return NULL;

    memset(toset,0,sizeof(fx_week));

    toset->days[0].dayname = "Monday";
    toset->days[1].dayname = "Tuesday";
    toset->days[2].dayname = "Wednesday";
    toset->days[3].dayname = "Thursday";
    toset->days[4].dayname = "Friday";

    toset->days[5].dayname = "Saturday";
    toset->days[6].dayname = "Sunday";

    toset->weekno=weeknumber;
    toset->yearno=yearnumber;
    toset->dayno=get_week_day_number();

    /* read master control file for week and calc normal hours */
    read_week_control(docroot,toset);

    for (loop=0;loop<7;loop++)
    {
	toset->days[loop].daytype=toset->days[loop].daytype_default;
    }

    toset->week_length=calc_normal_hours(toset);

    return toset;
}

int update_week_days(fx_day * day)
{
    char * cgi_entry_name;
    int loop;
    int type;
    int hour,min;

    cgi_entry_name=malloc(1024);

    for (loop=0;loop<INOUT_PER_DAY;loop++)
    {
	sprintf(cgi_entry_name,"%s_h_inout_%d",day->dayname,loop);
	hour=cgi_getentryint(cgi_entry_name);
	sprintf(cgi_entry_name,"%s_m_inout_%d",day->dayname,loop);
	min=cgi_getentryint(cgi_entry_name);

	/* no more than 24 hours in a day (0 ref)*/
	if(hour>(HOURS_PER_DAY-1))
	    hour=(HOURS_PER_DAY-1);

	if(!((hour==0) && (min==0)))
	    day->inout[loop]=((hour*SECONDS_PER_HOUR)+(min*SECONDS_PER_MIN));
    }

    /* check for daytype changes - only if global type is normal day*/
    sprintf(cgi_entry_name,"%s_daytype",day->dayname);
    type=cgi_getentryint(cgi_entry_name);
    if (type!=0)
    {
        if((day->daytype_default==DAYTYPE_WORKING) && (type>DAYTYPE_GLOBALEND))
            day->daytype = type;
        else
            day->daytype = day->daytype_default;
    }

    free(cgi_entry_name);

    /* ensure user input never goes back in time! */
    for (loop=1;loop<4;loop++)
	if(day->inout[loop-1] > day->inout[loop])
	    day->inout[loop]=day->inout[loop-1];


    return 0;

}

int calc_worked_hours(fx_week * updateweek)
{
    int loop,loop2;
    int day_tot_time=0;
    fx_day * day;

    updateweek->week_leave=0;
    updateweek->week_total=0;
    updateweek->week_flexi=0;

    /* for each day */
    for (loop=0;loop<7;loop++)
    {
	day=&updateweek->days[loop];
	/* check its a working day */
	switch(day->daytype)
	{
	case DAYTYPE_WORKING :
	case DAYTYPE_BANKHOL :
	case DAYTYPE_WEEKEND :
	case DAYTYPE_OVERTIME:
	    day_tot_time=0;

	    /* claculate each day total time */
	    for (loop2=0;loop2<4;loop2+=2)
		day_tot_time+=((day->inout[loop2+1]) - (day->inout[loop2]));

	    day->day_total=day_tot_time;
	    break;

	case DAYTYPE_HALF_FLEXIDAY :
	    updateweek->week_flexi+=(updateweek->work_day_length/2);
	case DAYTYPE_HALF_HOLIDAY :
	case DAYTYPE_HALF_SICK :
	    updateweek->week_leave+=(updateweek->work_day_length/2);
	    day_tot_time=0;

	    /* claculate half day total time */
	    day->day_total=((day->inout[1]) - (day->inout[0]));
	    break;

	case DAYTYPE_FLEXIDAY :
	    updateweek->week_flexi+=updateweek->work_day_length;
	case DAYTYPE_SICK :
	case DAYTYPE_HOLIDAY :
	case DAYTYPE_CONFERENCE :
	    updateweek->week_leave+=updateweek->work_day_length;
	    day->day_total=0;
	    break;
	}
	updateweek->week_total+=day->day_total;
    }

    return updateweek->week_total;
}


int update_week(fx_week * updateweek)
{
    int loop;

    /* get vars of the form <day>_[hm]_inout_<num> */
    for (loop=0;loop<7;loop++)
	update_week_days(&updateweek->days[loop]);

    updateweek->week_length=calc_normal_hours(updateweek);

    calc_worked_hours(updateweek);

    return 0;
}


int save_week(fx_week * saveweek,const char * docroot,const char * user)
{
    FILE * sheet_file;
    int loop;

    if (user==NULL)
	return 1;

    sheet_file=open_sheet(docroot,user,saveweek->yearno,saveweek->weekno,"w+");
    if (sheet_file==NULL)
    {
	printf("<!-- unable to write to sheet file -->\n");
	return 0;
    }

    for(loop=0;loop<7;loop++)
    {
	fprintf(sheet_file,"%ld %ld %ld %ld\n",
		saveweek->days[loop].inout[0],
		saveweek->days[loop].inout[1],
		saveweek->days[loop].inout[2],
		saveweek->days[loop].inout[3]);
    }

    fprintf(sheet_file,"%d %d %d %d %d %d %d\n",
	    saveweek->days[0].daytype,
	    saveweek->days[1].daytype,
	    saveweek->days[2].daytype,
	    saveweek->days[3].daytype,
	    saveweek->days[4].daytype,
	    saveweek->days[5].daytype,
	    saveweek->days[6].daytype);

    fclose(sheet_file);
    return 0;
}

fx_week* new_week(const char * docroot,const char * user,int week_num,int year_num)
{
    fx_week *newweek;

    if (user==NULL)
	return NULL;

    newweek=initweek(docroot,week_num,year_num);

    save_week(newweek,docroot,user);

    return(newweek);
}

fx_week* load_week(const char * docroot,const char * user,int week_num,int year_num)
{
    FILE * sheet_file;
    fx_week *newweek;
    int loop,scanned;
    long daytype[7];

    if (user==NULL)
	return NULL;

    newweek=initweek(docroot,week_num,year_num);

    sheet_file=open_sheet(docroot,user,year_num,week_num,"r");

    if (sheet_file==NULL)
    {
	free(newweek);
	return NULL;
    }

    for(loop=0;loop<7;loop++)
    {
	scanned=fscanf(sheet_file,"%ld %ld %ld %ld\n",
		       &newweek->days[loop].inout[0],
		       &newweek->days[loop].inout[1],
		       &newweek->days[loop].inout[2],
		       &newweek->days[loop].inout[3]);

	/* check we have the right number of inputs and clear any in error */
	if (scanned!=4)
	{
	    if (scanned<0)
		scanned=0;
	    while(scanned<4)
	    {
		newweek->days[loop].inout[scanned]=0;
		scanned++;
	    }
	}

    }

    /* get the daytypes (for hols etc) check they are valid before use */
    scanned=fscanf(sheet_file,"%ld %ld %ld %ld %ld %ld %ld\n",
		&daytype[0],
		&daytype[1],
		&daytype[2],
		&daytype[3],
		&daytype[4],
		&daytype[5],
		&daytype[6]);

    /* check we have the right number of inputs and clear any in error */
    if (scanned!=7)
    {
	if (scanned<0)
	    scanned=0;
	while(scanned<7)
	{
	    daytype[scanned]=0;
	    scanned++;
	}
    }

    for(loop=0;loop<7;loop++)
    {
	if(daytype[loop]>DAYTYPE_GLOBALEND)
	    newweek->days[loop].daytype=daytype[loop];
    }

    newweek->week_length=calc_normal_hours(newweek);

    fclose(sheet_file);

    return(newweek);

}


int flexi_show_sheet(char * docroot,const char * user,const char * password,int weekno,int yearno)
{
    fx_week *thisweek;

    if ((user==NULL) || (strlen(user)==0))
    {
	printf("<H2>Error user was empty</h2>");
	return 1;
    }

    if (weekno<0)
	weekno=get_week_number();

    if (yearno==0)
	yearno=get_year_number();

    /* load week file into struct or create a new one as apropriate */
    thisweek=load_week(docroot,user,weekno,yearno);
    if (thisweek==NULL)
	thisweek=new_week(docroot,user,weekno,yearno);

    /* keep any changes made and calc totals */
    update_week(thisweek);

    /* if the user updated the form ensure we save the updates */
    save_week(thisweek,docroot,user);

    /* choose the action to take */
    return action(thisweek,docroot,user,password,cgi_getentrystr("action"));


}


int main(int argc, char * argv[])
{
    int res;
    char * doc_root;
    int weekno;

    /* Initialize the CGI and send out an HTTP header: */

    res = cgi_init();

    printf("Content-type: text/html\n\n");


    /* Was there an error initializing the CGI??? */

    if (res != CGIERR_NONE)
    {
	printf("Error # %d: %s<p>\n", res, cgi_strerror(res));
	exit(0);
    }

    doc_root=getenv("DOCUMENT_ROOT");
    if (doc_root == NULL)
    {
	send_error("Document Root Not Found");
	exit(0);
    }

    weekno=cgi_getentryint("week");
    if((weekno==0) && (cgi_errno==CGIERR_NOT_INTEGER))
	weekno=-1;

    /* boilerplate top here */
    printf("<html>\n");


    if(flexi_check_passwd(doc_root,
			  cgi_getentrystr("user"),
			  cgi_getentrystr("password")))
    {
	flexi_show_sheet(doc_root,
			 cgi_getentrystr("user"),
			 cgi_getentrystr("password"),
			 weekno,
			 cgi_getentryint("year"));
    }
    else
	flexi_prompt_passwd(doc_root,cgi_getentrystr("user"));


    /* boilerplate bottom here */
    printf("</html>\n");

    cgi_quit();

    return(0);
}
